package edu.rochester.cs.quagents3.client;
import java.awt.Image;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.*;

import javax.sound.sampled.SourceDataLine;

//class which represents the front-end interface for Quagents 3
//used by both gui and quagent ai
public class Client extends Thread implements ClientInf{
	private Socket sock,eventsock;
	public static int SERVERPORT = 6000; //global port reference
	public static int EVENTPORT = 6001; //global eventport reference
	public long commandid = 0; //global id reference
	public Stack<Command> comhist; //command history
	private OutputStream sockout,evout;
	private Writer writer,evwriter;
	private ReaderThread rt,et;
	private int qid = -2; //unique id
	public static int BBOXLENGTH = 15;
	private Updater update;
	public String name = "Nameless";
	private QEventListener event;
	private CommandTerminationListener comlist;
	private volatile List<ProtocolZero> running; //List of currently executing p0 funcs - used by reader 
	
	//configable?
	public static double DEFAULTSPEED = .4;
	public static priority_t DEFAULTPRIORITY = priority_t.NOW;
	public static int DEFAULT_RF_TYPE = 1;
	public static double DEFAULTVOLUME = -1;
	public static boolean VERBOSE = false; //used for debugging
	public static boolean synchronous = false; 
	
	private boolean ready = false; //ready flag, determines whether auto-update or imperative command
	
	public void run(){}
	
	//construction automatically connects
	public Client(String ip) throws UnknownHostException, IOException{
		sock = new Socket(ip,SERVERPORT);//x then y
		eventsock = new Socket(ip,EVENTPORT);
		sockout = sock.getOutputStream();
		evout = eventsock.getOutputStream();
		writer = new OutputStreamWriter(sockout);
		evwriter = new OutputStreamWriter(evout);
		comhist = new Stack<Command>();
		running = new LinkedList<ProtocolZero>();
		rt = new ReaderThread(this,sock);
		et = new ReaderThread(this,eventsock);
		update = new Updater();
		
		//construct default eventhandler
		addEventListener(new DefaultQEventListener(this));
		addEventListener(new DefaultCommandTerminationListener());
	}
	
	public void terminate(){
		rt.stopASAP(); et.stopASAP();
		rt = et = null;
	}
	
	/**
	 * Executes a command. The command is started, and blocks until termination. <br/>
	 * execute() is equivalent to atomically calling start() followed by waitForTerminate().
	 * @param com
	 */
	public void execute(Command com){
		start(com);
		waitForTerminate(com);
	}
	
	//async exe
	public void start(Command com){
		com.setExecutor(this);
		comhist.push(com);
		if(!com.periodic)
			com.run();
		else
			update.insert(com);
	}
	
	public void waitForTerminate(Command com){
		synchronized(com){
			while(!com.hasTerminated()){
				try{com.wait();}
				catch(InterruptedException e){}
			}
		}
	}
	
	public boolean isQuagent(){
		return true;
	}
	
	public List<ProtocolZero> getRunningList(){
		return running;
	}
	
	public long incrementCommandID(){
		return ++commandid;
	}
	
	public void removeCommandFromRunning(ProtocolZero com){
		running.remove(com);
	}

	public void addCommandToRunning(ProtocolZero com){
		running.add(com);
	}
	
	public void notifyCommandTerminationListener(Command com){
		comlist.commandTerminated(com);
	}
	
	public int getQuagentID(){
		if(qid==-2)
			System.err.println("error: Quagent ID not yet set. Please call ready() before this function");
		return qid;
	}
	
	/**
	 * Used by the reader thread to set the quagent's id.
	 */
	public void setQuagentID(int qid){
		if(ready)
			return;
		else
			this.qid = qid; 
	}
	
	/**
	 * True if found in updater, else false.
	 * @param com
	 * @return
	 */
	public boolean stop(Command com){
		return update.remove(com)!=null;
	}
	
	//user might want this
	public Updater getUpdater(){return update;}
	
	//for external Command creations
	public Writer getWriter(){return writer;}
	
	
	//true if connection is still valid
	//use this before r/w
	public boolean validateSocket() throws IOException{
		if(!sock.isConnected()){
			sock.close();
			System.out.println("Connection lost. Socket closed");
			return false;
		}
		else
			return true;
	}
	
	public static char convertPriority(priority_t priority){
		if(priority==priority_t.NOW)return 'n';
		else if(priority==priority_t.LATER)return 't';
		else if(priority==priority_t.REPLACE)return 'r';
		else {System.err.println("error with priority conversion");
		return '\0';}
	}
	
	public void ready() throws IOException{
		update.run();
		if(VERBOSE)
			System.out.println("writing: ready\n");
		writer.write("ready\n"); writer.flush();
		rt.start();
		while(qid==-2){System.out.print("");}
		evwriter.write("identity " + qid); evwriter.flush();
		et.start();
		ready = true;
	}
		
	public void name(String name) throws IOException{
		if(!ready){
		writer.write("name " + name + "\n"); writer.flush();
		this.name = name;
		}
	}
	
	public void botfile(String filename) throws IOException {
		writer.write("botfile " + filename + "\n"); writer.flush();
	}


	public void team(team t) throws IOException {
		String te="";
		if(t==team.RED) te="red";
		else if(t==team.BLUE) te="blue";
		writer.write("team " + te + "\n"); writer.flush();
	}
	
	public void password(String password) throws IOException {
		writer.write("password " + password + "\n"); writer.flush();
	}
	public void skill(int skill) throws IOException {
		writer.write("skill " + skill + "\n"); writer.flush();
	}
	
	public void initialPosition(double x, double y, double z) throws IOException{
		writer.write("initialpos " + x + " " + y + " " + z + "\n"); writer.flush();
	}
	
	//moveby
	public MoveBy moveBy(double dir, double dist, double speed, int obstacles, priority_t priority){
		MoveBy com = new MoveBy(dir,dist,speed,obstacles,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveBy moveBy(double dir, double dist, double speed, int obstacles, priority_t priority, long period){
		MoveBy com = new MoveBy(dir,dist,speed,obstacles,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveBy moveBy(double dir, double dist, double speed, int obstacles){return moveBy(dir,dist,speed,obstacles,DEFAULTPRIORITY);}
	public MoveBy moveBy(double dir, double dist, double speed, int obstacles, long period){return moveBy(dir,dist,speed,obstacles,DEFAULTPRIORITY,period);}
	public MoveBy moveBy(double dir, double dist, double speed, priority_t priority){return moveBy(dir,dist,speed,0,priority);}
	public MoveBy moveBy(double dir, double dist, double speed, priority_t priority, long period){return moveBy(dir,dist,speed,0,priority,period);}
	public MoveBy moveBy(double dir, double dist, double speed){return moveBy(dir,dist,speed,0,DEFAULTPRIORITY);}
	public MoveBy moveBy(double dir, double dist, double speed, long period){return moveBy(dir,dist,speed,0,DEFAULTPRIORITY,period);}
	public MoveBy moveBy(double dist, double speed, priority_t priority){return moveBy(0,dist,speed,0,priority);}
	public MoveBy moveBy(double dist, double speed, priority_t priority, long period){return moveBy(0,dist,speed,0,priority,period);}
	public MoveBy moveBy(double dist, double speed){return moveBy(0,dist,speed,0,DEFAULTPRIORITY);}
	public MoveBy moveBy(double dist, double speed, long period){return moveBy(0,dist,speed,0,DEFAULTPRIORITY,period);}
	public MoveBy moveBy(double dist, priority_t priority){return moveBy(0,dist,DEFAULTSPEED,0,priority);}
	public MoveBy moveBy(double dist, priority_t priority,long period){return moveBy(0,dist,DEFAULTSPEED,0,priority,period);}
	public MoveBy moveBy(double dist){return moveBy(0,dist,DEFAULTSPEED,0,DEFAULTPRIORITY);}
	public MoveBy moveBy(double dist, long period){return moveBy(0,dist,DEFAULTSPEED,0,DEFAULTPRIORITY,period);}
	
	public MoveTo moveTo(double[] point, double speed, int obstacles, priority_t priority){
		MoveTo com = new MoveTo(point,speed,obstacles,priority); //push onto command history
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveTo moveTo(double[] point, double speed, int obstacles, priority_t priority, long period){
		MoveTo com = new MoveTo(point,speed,obstacles,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveTo moveTo(double[] point, double speed, int obstacles){return moveTo(point,speed,obstacles,DEFAULTPRIORITY);}
	public MoveTo moveTo(double[] point, double speed, int obstacles, long period){return moveTo(point,speed,obstacles,DEFAULTPRIORITY,period);}
	public MoveTo moveTo(double[] point, double speed, priority_t priority){return moveTo(point,speed,0,priority);}
	public MoveTo moveTo(double[] point, double speed, priority_t priority, long period){return moveTo(point,speed,0,priority,period);}
	public MoveTo moveTo(double[] point, double speed){return moveTo(point,speed,0,DEFAULTPRIORITY);}
	public MoveTo moveTo(double[] point, double speed, long period){return moveTo(point,speed,0,DEFAULTPRIORITY,period);}
	public MoveTo moveTo(double[] point, priority_t priority){return moveTo(point,DEFAULTSPEED,0,priority);}
	public MoveTo moveTo(double[] point, priority_t priority, long period){return moveTo(point,DEFAULTSPEED,0,priority, period);}
	public MoveTo moveTo(double[] point){return moveTo(point,DEFAULTSPEED,0,DEFAULTPRIORITY);}
	public MoveTo moveTo(double[] point, long period){return moveTo(point,DEFAULTSPEED,0,DEFAULTPRIORITY,period);}
	
	//moveforever
	public MoveForever moveForever(double dir, double speed, int obstacles, priority_t priority){
		MoveForever com = new MoveForever(dir,speed,obstacles,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveForever moveForever(double dir, double speed, int obstacles, priority_t priority, long period){
		MoveForever com = new MoveForever(dir,speed,obstacles,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveForever moveForever(double dir, double speed, int obstacles){return moveForever(dir,speed,obstacles,DEFAULTPRIORITY);}
	public MoveForever moveForever(double dir, double speed, int obstacles, long period){return moveForever(dir,speed,obstacles,DEFAULTPRIORITY,period);}
	public MoveForever moveForever(double dir, double speed, priority_t priority){return moveForever(dir,speed,0,priority);}
	public MoveForever moveForever(double dir, double speed, priority_t priority, long period){return moveForever(dir,speed,0,priority,period);}
	public MoveForever moveForever(double dir, double speed){return moveForever(dir,speed,0,DEFAULTPRIORITY);}
	public MoveForever moveForever(double dir, double speed, long period){return moveForever(dir,speed,0,DEFAULTPRIORITY,period);}
	public MoveForever moveForever(double speed, priority_t priority){return moveForever(0,speed,0,priority);}
	public MoveForever moveForever(double speed, priority_t priority, long period){return moveForever(0,speed,0,priority,period);}
	public MoveForever moveForever(double speed){return moveForever(0,speed,0,DEFAULTPRIORITY);}
	public MoveForever moveForever(double speed, long period){return moveForever(0,speed,0,DEFAULTPRIORITY,period);}
	public MoveForever moveForever(priority_t priority){return moveForever(DEFAULTSPEED, priority);}
	public MoveForever moveForever(priority_t priority, long period){return moveForever(DEFAULTSPEED, priority,period);}
	public MoveForever moveForever(){return moveForever(DEFAULTPRIORITY);}
	public MoveForever moveForever(long period){return moveForever(DEFAULTPRIORITY,period);}
	
	public MoveFor moveFor(double dir, double time, double speed, int obstacles, priority_t priority){
		MoveFor com = new MoveFor(dir,time,speed,obstacles,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveFor moveFor(double dir, double time, double speed, int obstacles, priority_t priority, long period){
		MoveFor com = new MoveFor(dir,time,speed,obstacles,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MoveFor moveFor(double dir, double time, double speed, int obstacles){return moveFor(dir,time,speed,obstacles,DEFAULTPRIORITY);}
	public MoveFor moveFor(double dir, double time, double speed, int obstacles, long period){return moveFor(dir,time,speed,obstacles,DEFAULTPRIORITY, period);}
	public MoveFor moveFor(double dir, double time, double speed, priority_t priority){return moveFor(dir,time,speed,0,priority);}
	public MoveFor moveFor(double dir, double time, double speed, priority_t priority, long period){return moveFor(dir,time,speed,0,priority, period);}
	public MoveFor moveFor(double dir, double time, double speed){return moveFor(dir,time,speed,0,DEFAULTPRIORITY);}
	public MoveFor moveFor(double dir, double time, double speed, long period){return moveFor(dir,time,speed,0,DEFAULTPRIORITY, period);}
	public MoveFor moveFor(double time, double speed, priority_t priority){return moveFor(0,time,speed,0,priority);}
	public MoveFor moveFor(double time, double speed, priority_t priority, long period){return moveFor(0,time,speed,0,priority, period);}
	public MoveFor moveFor(double time, double speed){return moveFor(0,time,speed,0,DEFAULTPRIORITY);}
	public MoveFor moveFor(double time, double speed, long period){return moveFor(0,time,speed,0,DEFAULTPRIORITY, period);}
	public MoveFor moveFor(double time, priority_t priority){return moveFor(0,time,DEFAULTSPEED,0,priority);}
	public MoveFor moveFor(double time, priority_t priority, long period){return moveFor(0,time,DEFAULTSPEED,0,priority, period);}
	public MoveFor moveFor(double time){return moveFor(0,time,DEFAULTSPEED,0,DEFAULTPRIORITY);}
	public MoveFor moveFor(double time, long period){return moveFor(0,time,DEFAULTSPEED,0,DEFAULTPRIORITY,period);}
	
	public Jump jump(double dir, double speed, priority_t priority){
		Jump com = new Jump(dir,speed,priority);		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Jump jump(double dir, double speed, priority_t priority, long period){
		Jump com = new Jump(dir,speed,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Jump jump(double dir, double speed){return jump(dir,speed,DEFAULTPRIORITY);}
	public Jump jump(double dir, double speed, long period){return jump(dir,speed,DEFAULTPRIORITY, period);}
	public Jump jump(priority_t priority){return jump(0,DEFAULTSPEED,priority);}
	public Jump jump(priority_t priority, long period){return jump(0,DEFAULTSPEED,priority,period);}
	public Jump jump(){return jump(0,DEFAULTSPEED,DEFAULTPRIORITY);}
	public Jump jump(long period){return jump(0,DEFAULTSPEED,DEFAULTPRIORITY,period);}
	

	public CurrentArmor currentArmor(priority_t priority){
		CurrentArmor com = new CurrentArmor(priority); 
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentArmor currentArmor(priority_t priority, long period){
		CurrentArmor com = new CurrentArmor(priority,period); //push onto command history
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentArmor currentArmor(){return currentArmor(DEFAULTPRIORITY);}
	public CurrentArmor currentArmor(long period){return currentArmor(DEFAULTPRIORITY, period);}

	public CurrentHealth currentHealth(priority_t priority){
		CurrentHealth com = new CurrentHealth(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentHealth currentHealth(priority_t priority, long period){
		CurrentHealth com = new CurrentHealth(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentHealth currentHealth(){return currentHealth(DEFAULTPRIORITY);}
	public CurrentHealth currentHealth(long period){return currentHealth(DEFAULTPRIORITY,period);}

	public CurrentAmmo currentAmmo(weapon wep, priority_t priority){
		CurrentAmmo com = new CurrentAmmo(wep,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentAmmo currentAmmo(weapon wep, priority_t priority, long period){
		CurrentAmmo com = new CurrentAmmo(wep,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentAmmo currentAmmo(weapon wep){return currentAmmo(wep,DEFAULTPRIORITY);}
	public CurrentAmmo currentAmmo(weapon wep, long period){return currentAmmo(wep,DEFAULTPRIORITY,period);}
	
	
	public FireWeapon fireWeapon(priority_t priority){
		FireWeapon com = new FireWeapon(priority); //push onto command history
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public FireWeapon fireWeapon(){return fireWeapon(DEFAULTPRIORITY);}
	public FireWeapon fireWeapon(priority_t priority, long period){
		FireWeapon com = new FireWeapon(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public FireWeapon fireWeapon(long period){return fireWeapon(DEFAULTPRIORITY,period);}

	public Skip skip(priority_t priority){
		Skip com = new Skip(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Skip skip(){return skip(DEFAULTPRIORITY);}
	public Skip skip(priority_t priority, long period){
		Skip com = new Skip(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Skip skip(long period){return skip(DEFAULTPRIORITY,period);}
	
	public ForgetAllTasks forgetAllTasks(priority_t priority) {
		ForgetAllTasks com = new ForgetAllTasks(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public ForgetAllTasks forgetAllTasks(){return forgetAllTasks(DEFAULTPRIORITY);}
	public ForgetAllTasks forgetAllTasks(priority_t priority, long period) {
		ForgetAllTasks com = new ForgetAllTasks(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public ForgetAllTasks forgetAllTasks(long period){return forgetAllTasks(DEFAULTPRIORITY,period);}

	
	public Rangefinder rangefinder(int type, int range, double rot, double azi, priority_t priority){
		Rangefinder com = new Rangefinder(type,range,rot,azi,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Rangefinder rangefinder(int type, int range, double rot, double azi, priority_t priority, long period){
		Rangefinder com = new Rangefinder(type,range,rot,azi,priority,period); 
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Rangefinder rangefinder(int type, int range, double rot, double azi, long period){return rangefinder(type,range,rot,azi,DEFAULTPRIORITY,period);}
	public Rangefinder rangefinder(int type, int range, double rot, double azi){return rangefinder(type,range,rot,azi,DEFAULTPRIORITY);}
	public Rangefinder rangefinder(int type, double rot, double azi, priority_t priority){return rangefinder(type,9999,rot,azi,priority);}
	public Rangefinder rangefinder(int type, double rot, double azi, priority_t priority, long period){return rangefinder(type,9999,rot,azi,priority, period);}
	public Rangefinder rangefinder(int type, double rot, double azi){return rangefinder(type,9999,rot,azi,DEFAULTPRIORITY);}
	public Rangefinder rangefinder(int type, double rot, double azi, long period){return rangefinder(type,9999,rot,azi,DEFAULTPRIORITY,period);}
	public Rangefinder rangefinder(int type, double rot, priority_t priority){return rangefinder(type,9999,rot,0,priority);}
	public Rangefinder rangefinder(int type, double rot, priority_t priority, long period){return rangefinder(type,9999,rot,0,priority,period);}
	public Rangefinder rangefinder(int type, double rot){return rangefinder(type,9999,rot,0,DEFAULTPRIORITY);}
	public Rangefinder rangefinder(int type, double rot, long period){return rangefinder(type,9999,rot,0,DEFAULTPRIORITY,period);}
	public Rangefinder rangefinder(double rot, priority_t priority){return rangefinder(DEFAULT_RF_TYPE,9999,rot,0,priority);}
	public Rangefinder rangefinder(double rot, priority_t priority, long period){return rangefinder(DEFAULT_RF_TYPE,9999,rot,0,priority,period);}
	public Rangefinder rangefinder(double rot){return rangefinder(rot,DEFAULTPRIORITY);}
	public Rangefinder rangefinder(double rot,long period){return rangefinder(rot,DEFAULTPRIORITY,period);}
	public Rangefinder rangefinder(int type, priority_t priority){return rangefinder(type,9999,0,0,priority);}
	public Rangefinder rangefinder(int type, priority_t priority, long period){return rangefinder(type,9999,0,0,priority,period);}
	public Rangefinder rangefinder(int type){return rangefinder(type,9999,0,0,DEFAULTPRIORITY);}
	public Rangefinder rangefinder(int type, long period){return rangefinder(type,9999,0,0,DEFAULTPRIORITY,period);}
	public Rangefinder rangefinder(priority_t priority){return rangefinder(1,9999,0,0,priority);}
	public Rangefinder rangefinder(priority_t priority, long period){return rangefinder(1,9999,0,0,priority,period);}
	public Rangefinder rangefinder(){return rangefinder(DEFAULTPRIORITY);}
	public Rangefinder rangefinder(long period){return rangefinder(DEFAULTPRIORITY,period);}
	
	public Say say(String msg, double volume, priority_t priority){
		Say com = new Say(volume,msg,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Say say(String msg, double volume, priority_t priority, long period){
		Say com = new Say(volume,msg,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Say say(String msg, double volume){return say(msg,volume,DEFAULTPRIORITY);}
	public Say say(String msg, double volume, long period){return say(msg,volume,DEFAULTPRIORITY,period);}
	public Say say(String msg, priority_t priority){return say(msg,DEFAULTVOLUME,priority);}
	public Say say(String msg, priority_t priority, long period){return say(msg,DEFAULTVOLUME,priority,period);}
	public Say say(String msg){return say(msg,DEFAULTVOLUME,DEFAULTPRIORITY);}
	public Say say(String msg, long period){return say(msg,DEFAULTVOLUME,DEFAULTPRIORITY,period);} 
	
	public CurrentLocation currentLocation(priority_t priority){
		CurrentLocation com = new CurrentLocation(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentLocation currentLocation(){return currentLocation(DEFAULTPRIORITY);}	
	public CurrentLocation currentLocation(priority_t priority, long period){
		CurrentLocation com = new CurrentLocation(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CurrentLocation currentLocation(long period){return currentLocation(DEFAULTPRIORITY,period);}


	public MaxArmor maxArmor(priority_t priority){
		MaxArmor com = new MaxArmor(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MaxArmor maxArmor(){return maxArmor(DEFAULTPRIORITY);}
	public MaxArmor maxArmor(priority_t priority, long period){
		MaxArmor com = new MaxArmor(priority,period); 
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MaxArmor maxArmor(long period){return maxArmor(DEFAULTPRIORITY,period);}
	
	public MaxHealth maxHealth(priority_t priority){
		MaxHealth com = new MaxHealth(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MaxHealth maxHealth(){return maxHealth(DEFAULTPRIORITY);}
	public MaxHealth maxHealth(priority_t priority, long period){
		MaxHealth com = new MaxHealth(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public MaxHealth maxHealth(long period){return maxHealth(DEFAULTPRIORITY,period);}

	public CanSee canSee(int entityid, priority_t priority){
		CanSee com = new CanSee(entityid,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CanSee canSee(int entityid){return canSee(entityid, DEFAULTPRIORITY);}
	public CanSee canSee(int entityid, priority_t priority, long period){
		CanSee com = new CanSee(entityid,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CanSee canSee(int entityid, long period){return canSee(entityid, DEFAULTPRIORITY,period);}
	
	public Pause pause(priority_t priority){
		Pause com = new Pause(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Pause pause(){return pause(DEFAULTPRIORITY);}
	public Pause pause(priority_t priority, long period){
		Pause com = new Pause(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Pause pause(long period){return pause(DEFAULTPRIORITY,period);}
	
	public Pop pop(priority_t priority){
		Pop com = new Pop(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Pop pop(){return pop(DEFAULTPRIORITY);}
	public Pop pop(priority_t priority,long period){
		Pop com = new Pop(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Pop pop(long period){return pop(DEFAULTPRIORITY,period);}
	
	public Peek peek(int depth, priority_t priority){
		Peek com = new Peek(depth, priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Peek peek(int depth){return peek(depth,DEFAULTPRIORITY);}
	public Peek peek(priority_t priority){return peek(0,priority);}
	public Peek peek(){return peek(0,DEFAULTPRIORITY);}
	public Peek peek(int depth, priority_t priority, long period){
		Peek com = new Peek(depth, priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Peek peek(int depth,long period){return peek(depth,DEFAULTPRIORITY,period);}
	public Peek peek(priority_t priority,long period){return peek(0,priority,period);}
	public Peek peek(long period){return peek(0,DEFAULTPRIORITY,period);}
	
	
	public Rotate rotate(double rot, double azi, priority_t priority){
		Rotate com = new Rotate(rot,azi,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Rotate rotate(double rot, double azi){return rotate(rot,azi,DEFAULTPRIORITY);}
	public Rotate rotate(double rot, priority_t priority){return rotate(rot,0,priority);}
	public Rotate rotate(double rot){return rotate(rot,0,DEFAULTPRIORITY);}
	public Rotate rotate(double rot, double azi, priority_t priority, long period){
		Rotate com = new Rotate(rot,azi,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Rotate rotate(double rot, double azi, long period){return rotate(rot,azi,DEFAULTPRIORITY,period);}
	public Rotate rotate(double rot, priority_t priority, long period){return rotate(rot,0,priority,period);}
	public Rotate rotate(double rot, long period){return rotate(rot,0,DEFAULTPRIORITY,period);}
	

	public Facing facing(priority_t priority){
		Facing com = new Facing(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Facing facing(){return facing(DEFAULTPRIORITY);}
	public Facing facing(priority_t priority, long period){
		Facing com = new Facing(priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Facing facing(long period){return facing(DEFAULTPRIORITY,period);}

	
	public Crouch crouch(boolean tocrouch, priority_t priority) {
		Crouch com = new Crouch(tocrouch,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Crouch crouch(boolean tocrouch) {return crouch(tocrouch,DEFAULTPRIORITY);}
	public Crouch crouch(boolean tocrouch, priority_t priority, long period) {
		Crouch com = new Crouch(tocrouch,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Crouch crouch(boolean tocrouch, long period) {return crouch(tocrouch,DEFAULTPRIORITY,period);}

	public SwitchWeapon switchWeapon(weapon wep, priority_t priority) {
		SwitchWeapon com = new SwitchWeapon(wep,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public SwitchWeapon switchWeapon(weapon wep) {return switchWeapon(wep,DEFAULTPRIORITY);}
	public SwitchWeapon switchWeapon(weapon wep, priority_t priority, long period) {
		SwitchWeapon com = new SwitchWeapon(wep,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public SwitchWeapon switchWeapon(weapon wep, long period) {return switchWeapon(wep,DEFAULTPRIORITY,period);}
	

	public Shove shove(double strength, double yaw, double pitch, priority_t priority) {
		Shove com = new Shove(strength,yaw,pitch,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Shove shove(double strength, double yaw, double pitch, priority_t priority, long period) {
		Shove com = new Shove(strength,yaw,pitch,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Shove shove(double strength, double yaw, double pitch) {return shove(strength,yaw,pitch,DEFAULTPRIORITY);}
	public Shove shove(double strength, double yaw, double pitch, long period) {return shove(strength,yaw,pitch,DEFAULTPRIORITY,period);}
	public Shove shove(double strength, double yaw, priority_t priority){return shove(strength,yaw,0,priority);}
	public Shove shove(double strength, double yaw, priority_t priority,long period) {return shove(strength,yaw,0,priority,period);}
	public Shove shove(double strength, double yaw) {return shove(strength,yaw,0);}
	public Shove shove(double strength, double yaw, long period) {return shove(strength,yaw,0,period);}
	public Shove shove(double strength, priority_t priority) {return shove(strength,0,priority);}
	public Shove shove(double strength, priority_t priority, long period) {return shove(strength,0,0,priority,period);}
	public Shove shove(double strength) {return shove(strength,DEFAULTPRIORITY);}
	public Shove shove(double strength, long period) {return shove(strength,DEFAULTPRIORITY,period);}
	
	
	public DepthViewpoint depthViewpoint(priority_t priority){
		DepthViewpoint com = new DepthViewpoint(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public DepthViewpoint depthViewpoint(){return depthViewpoint(DEFAULTPRIORITY);}
	public DepthViewpoint depthViewpoint(long period){
		DepthViewpoint com = new DepthViewpoint(period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public DepthViewpoint depthViewpoint(long period, priority_t priority){
		DepthViewpoint com = new DepthViewpoint(period, priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	
	public ImageViewpoint imageViewpoint(priority_t priority){
		ImageViewpoint com = new ImageViewpoint(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public ImageViewpoint imageViewpoint(){return imageViewpoint(DEFAULTPRIORITY);} 
	public ImageViewpoint imageViewpoint(long period){
		ImageViewpoint com = new ImageViewpoint(period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public ImageViewpoint imageViewpoint(long period, priority_t priority){
		ImageViewpoint com = new ImageViewpoint(period,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	
	public PickUp pickUp(long period, priority_t priority){
		PickUp com = new PickUp(period,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public PickUp pickUp(long period){
		PickUp com = new PickUp(period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public PickUp pickUp(priority_t priority){
		PickUp com = new PickUp(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public PickUp pickUp(){
		PickUp com = new PickUp();
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	
	public PutDown putDown(int item, int quantity){
		PutDown com = new PutDown(item, quantity);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public PutDown putDown(int item, int quantity, long period){
		PutDown com = new PutDown(item, quantity, period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public PutDown putDown(int item, int quantity, priority_t priority){
		PutDown com = new PutDown(item, quantity, priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public PutDown putDown(int item, int quantity, long period, priority_t priority){
		PutDown com = new PutDown(item, quantity, period, priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	
	
	public CheckInventory checkInventory(long period, priority_t priority){
		CheckInventory com = new CheckInventory(period,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CheckInventory checkInventory(priority_t priority){
		CheckInventory com = new CheckInventory(priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CheckInventory checkInventory(long period){
		CheckInventory com = new CheckInventory(period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public CheckInventory checkInventory(){
		CheckInventory com = new CheckInventory();
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	
	//p1
	public RotateFireWeapon rotateFireWeapon(double rot, double azi){
		RotateFireWeapon com = new RotateFireWeapon(rot,azi);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}	
	
    public Follow follow(int entityid, priority_t priority) {
    	Follow com = new Follow(entityid);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
    }
    public Follow follow(int entityid){return follow(entityid, DEFAULTPRIORITY);}
   //todo: fix this
    public Follow follow(int entityid, long duration, priority_t priority){
    	/*Follow com = new Follow(entityid, duration, priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;*/
    	return null;
    }
    public Follow follow(int entityid, long duration){return follow(entityid,duration,DEFAULTPRIORITY);}

	public Radar radar(double range, priority_t priority) {
		Radar com = new Radar(range,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Radar radar(double range, priority_t priority, long period) {
		Radar com = new Radar(range,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public Radar radar(double range){return radar(range, DEFAULTPRIORITY);}
	public Radar radar(double range, long period){return radar(range, DEFAULTPRIORITY,period);}
	public Radar radar(priority_t priority){return radar(9999,priority);}
	public Radar radar(priority_t priority,long period){return radar(9999,priority,period);}
	public Radar radar(){return radar(9999,DEFAULTPRIORITY);}
	public Radar radar(long period){return radar(9999,DEFAULTPRIORITY,period);}

	public WhatIs whatIs(int entityid, priority_t priority) {
		WhatIs com = new WhatIs(entityid,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public WhatIs whatIs(int entityid){return whatIs(entityid, DEFAULTPRIORITY);}
	public WhatIs whatIs(int entityid, priority_t priority, long period) {
		WhatIs com = new WhatIs(entityid,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public WhatIs whatIs(int entityid, long period){return whatIs(entityid, DEFAULTPRIORITY,period);}

	public BatchRangefinder batchRangefinder(int type, int distance, double rot,
			double azi, int width, int height, double sampleDeltaYaw,
			double sampleDeltaPitch,priority_t priority){
		BatchRangefinder com = new BatchRangefinder(type,distance,rot,azi,width,height,
				sampleDeltaYaw,sampleDeltaPitch,priority);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public BatchRangefinder batchRangefinder(int type, int distance, double rot,
			double azi, int width, int height, double sampleDeltaYaw,
			double sampleDeltaPitch,priority_t priority, long period){
		BatchRangefinder com = new BatchRangefinder(type,distance,rot,azi,width,height,
				sampleDeltaYaw,sampleDeltaPitch,priority,period);
		if(synchronous)
			execute(com);
		else
			start(com);
		return com;
	}
	public BatchRangefinder batchRangefinder(int type, int distance, double rot, double azi, int width,
			int height, double sampleDeltaYaw,double sampleDeltaPitch){
		return batchRangefinder(type,distance,rot,azi,width,height,sampleDeltaYaw,sampleDeltaPitch,DEFAULTPRIORITY);}
	public BatchRangefinder batchRangefinder(int type, int distance, double rot, double azi, int width,
			int height, double sampleDeltaYaw,double sampleDeltaPitch,long period){
		return batchRangefinder(type,distance,rot,azi,width,height,sampleDeltaYaw,sampleDeltaPitch,DEFAULTPRIORITY,period);}
	public BatchRangefinder batchRangefinder(int type, double rot, double azi, int width, int height,
			double sampleDeltaYaw,double sampleDeltaPitch, priority_t priority){
		return batchRangefinder(type,9999,rot,azi,width,height,sampleDeltaYaw,sampleDeltaPitch,priority);}
	public BatchRangefinder batchRangefinder(int type, double rot, double azi, int width, int height,
			double sampleDeltaYaw,double sampleDeltaPitch, priority_t priority,long period){
		return batchRangefinder(type,9999,rot,azi,width,height,sampleDeltaYaw,sampleDeltaPitch,priority,period);}
	public BatchRangefinder batchRangefinder(int type, double rot, double azi, int width, int height,
			double sampleDeltaYaw,double sampleDeltaPitch){
		return batchRangefinder(type,9999,rot,azi,width,height,sampleDeltaYaw,sampleDeltaPitch,DEFAULTPRIORITY);}
	public BatchRangefinder batchRangefinder(int type, double rot, double azi, int width, int height,
			double sampleDeltaYaw,double sampleDeltaPitch,long period){
		return batchRangefinder(type,9999,rot,azi,width,height,sampleDeltaYaw,sampleDeltaPitch,DEFAULTPRIORITY,period);}
	
	public void moveToObject(String object, int radius, priority_t priority) {
		Radar ra = this.radar(radius, priority);
		ra.waitForTerminate();
		int length = ra.getNumberEntitiesFound();
		Facing fa = this.facing();
		CurrentLocation cl = this.currentLocation();
		double location[] = cl.getCurrentLocation();
		double quagentyaw = fa.getViewDirection()[0];
		double quagentpitch = fa.getViewDirection()[1];
		for (int i = 0; i < length; i ++){
			if (ra.getClasstype(i).equals(object)){
				double temp[] = new double[2];
				temp[0] = location[0];
				temp[1] = location[1];
				double distance = ra.getDistance(i);
				distance *= Math.cos((quagentpitch + ra.getFacing(i)[1]) / 360 * 2 * Math.PI);
				temp[0] += distance*Math.cos((quagentyaw + ra.getFacing(i)[0]) / 360 * 2 * Math.PI);
				temp[1] += distance*Math.sin((quagentyaw + ra.getFacing(i)[0]) / 360 * 2 * Math.PI);
				MoveTo mt = this.moveTo(temp, 0.4, priority_t.LATER);
				mt.waitForTerminate();
			}
		}
	}
	public void moveToObject(String object, int radius){moveToObject(object,radius,DEFAULTPRIORITY);}
	public void moveToObject(String object, priority_t priority){moveToObject(object,9999,priority);}
	public void moveToObject(String object){moveToObject(object,9999,DEFAULTPRIORITY);}
	
	public void addEventListener(QEventListener qcp){
		this.event = qcp;
	}
	
	public QEventListener getEventListener(){
		return event;
	}
	
	public void addEventListener(CommandTerminationListener list){
		this.comlist = list;
	}
	
	private class DefaultQEventListener implements QEventListener{
		QCP protocol;
		
		public DefaultQEventListener(Client client){
			protocol = new QCP(client);
		}
		
		public void messageHeard(QEvent event) {
			System.out.println("Quagent " + event.getOrigin() + " says \""+event.getEvent()+"\""); 
			System.out.flush();
			protocol.parseMessage(event.getEvent(), event.getOrigin());
		}
		public void hasDied(QEvent event) {
			System.out.println("Quagent " + event.getOrigin() + " has died.");
		}
	}
	
	//don't need this anymore, use termination listener in Command.java
	public class DefaultCommandTerminationListener implements CommandTerminationListener{public void commandTerminated(Command com) {}}
}
	
