package edu.rochester.cs.quagents3.client;
import java.util.Scanner;
import java.util.TimerTask;

/**
 * Superclass for all commands. Data is unique to each command, so the parsing and storage of data is left to subclasses.
 */
abstract public class Command extends TimerTask implements Comparable<Command> {
		
	protected String op; //immutable
	protected long id; //immutable
	protected long inittime;
	protected volatile boolean running = false;
	protected long finaltime;
	protected boolean isread;
	protected long delay;
	protected long period;
	protected final CommandType whatami;
	protected String datastring;
	protected Entity executor = null;
	public final boolean periodic;
	protected CommandTerminationListener comlist;
	protected boolean paused = false;
	
	protected Command(boolean isread,CommandType whatami){
		this.isread = isread;
		this.whatami=whatami;
		periodic = false;
		addTerminationListener(new DefaultCommandTerminationListener());
	}
	
	protected Command(boolean isread,CommandType whatami, long period){
		this.isread = isread;
		this.whatami=whatami;
		this.period=period;
		periodic = true;
		addTerminationListener(new DefaultCommandTerminationListener());
	}
	
	protected Command(boolean isread,CommandType whatami, long period, long delay){
		this.isread = isread;
		this.whatami=whatami;
		this.period=period;
		this.delay=delay;
		periodic = true;
		addTerminationListener(new DefaultCommandTerminationListener());
	}
	
	
/**
 * Pauses a periodic command's execution. Since there is no support in TimerTask
 * or Timer for this, we will just use a flag on run().
 */
abstract public void pause();
	
abstract public void resume();
	
public void addTerminationListener(CommandTerminationListener comlist){
	this.comlist = comlist;
}
	
//blocks for server return
//updates data
//useful for sensors
abstract public void update();
	
//does nothing if no client attached
public Command waitForTerminate(){
	if(executor!=null)
		executor.waitForTerminate(this);
	return this;
}

abstract public void setExecutor(Entity c);

/**
 * Allows access to unique identifier. Used to identify command types during polymorphic uses.
 * @return String command type identifier.
 */
public CommandType whatAmI(){return whatami;}

/**
 * Used for scheduled commands. Returns delay before execution.
 * @return Delay before start of fixed-rate execution.
 * @usage Doesn't seem too useful...but user may need at some point.
 */
public long getDelay(){return delay;}

/**
 * Used for scheduled commands. Returns period of execution.
 * @return Period of fixed-rate execution.
 */
public long getPeriod(){return period;}

/**
 * Used for scheduled commands. Returns frequency of execution.
 * @return Frequency of fixed-rate execution.
 */
public double getFrequency(){return 1/period;}

/**
 * Returns op string sent to server. 
 * @return Formatted instruction string sent to server.
 */
public String getOp(){return op;} //usually isn't too useful for user

/**
 * Function which returns the current runtime of the command.
 * @return Current runtime of this command.
 * @usage For in-progress runtime update.
 */
public long getRuntime(){return System.currentTimeMillis()-inittime;}

/**
 * External running check.
 * @return True if command has terminated, else false.
 */
abstract public boolean hasTerminated();

//true if client is verified
protected boolean verifyClient(){
	if(executor==null){
		System.err.println("Error: Must specifiy executor before command execution");
		return false;
	}
	return true;
}

/**
 * Command id fetch function.
 * @return This command's unique command id.
 */
public long getID(){return id;}

/**
 * This function spins the calling thread until the command is terminated. Although all data functions block automatically,
 * this might be useful to the user anyways.
 * @bug Straightforward spinning didn't work for some reason (probably concurrent access bug), uses a hack for now.
 * @usage Whenever the user wants to wait for the command to terminate before proceeding.
 */
//public void waitForTerminate(){while(running){System.out.print("");}} //this is a [poor] hack...needs fixing

String ecode = "";
/**
 * Terminates the command, letting unlocking all waiting data fetchers.
 * Performs termination procedures and notifies client of completion. 
 * This command should <b>only</b> be called after setting all command data
 * using setData().
 * @param exitstring Exitcode for command.
 */

//TODO: improve verbose output
public String toString(){
	String rtn = "";
	if(running){
		rtn += "-------------------------------------------------------\n"+
			//+ "Quagent: " + executor.getQuagentID() + ",\t"+executor.name+ 
			"\nCommand: " + id + " - " + whatami +
			//"\nStarted at " + (short)inittime+
			"\nIn progress"+
			"\n-------------------------------------------------------";
	}
	else{
			rtn+="-------------------------------------------------------\n"+
			//	+ "Quagent: " + executor.getQuagentID() + ",\t"+executor.name+ 
				"\nCommand: " + id + " - " + whatami + 
			//	"\nStarted at " + (short) inittime +
				"\nExited with exit code " + ecode+'\n';
			if(isread) rtn+="Returned the value(s): " + toDataString() +"\n";
			rtn+="Execution time: " + (short)finaltime + "ms" +
				"\n-------------------------------------------------------";
	}
	return rtn;
}

/**
 * Returns raw data from server. Used by toString().
 * @return Raw data string.
 */
public String toDataString(){
	while(running){}
	return datastring;
}

/**
 * Used to access command exitcode.
 * @return Command exitcode.
 */
public String getExitCode(){return ecode;}

@Override
abstract public void run();

//used by scheduler
public int compareTo(Command comp) {
	//first check class vars
	/*if(this.isread!=comp.isread)
		return -1;
	else if(this.delay!=comp.delay) //is this necesary?
		return -1;
	else if(this.period!=comp.period)
		return -1;
	/*Scanner scan = new Scanner(getOp());
	Scanner scan2 = new Scanner(comp.getOp());
	if(!scan.next().equals(scan2.next())) //priority
		return -1;
	if(!scan.next().equals(scan2.next())) //opcode
		return -1;
	scan.next(); scan2.next(); //burn id, this doesn't matter for equality
	while(scan.hasNext()||scan2.hasNext()){
		if(!scan.hasNext()&&scan2.hasNext()) return 1; //need same number of params
		if(!scan.next().equals(scan.next())) return 1; //need same params
	}*/
	if(comp.id==this.id)
		return 0;
	else
		return -1;
}

private class DefaultCommandTerminationListener implements CommandTerminationListener{

	public void commandTerminated(Command com) {
		if(Client.VERBOSE)
			System.out.println(com.toString());
	}

}
}
