package edu.rochester.cs.quagents3.client;

import java.util.*;

/**
 * Superclass for all ProtocolOne commands. P1 doesn't care about priority. <br/>
 * This is made to be highly extensible. Developers who wish to add their own custom compound
 * functions to the Quagents standard library should extend from here.
 * <br/><br/> <b>In order to create a new compound function class, the class must define three functions: </b>
 * <ul>
 * <li>A <i>run()</i> function. This should start executing any action code. </li>
 * <li>An <i>update()</i> function. This should block until the function has
 * received a termination reply from the server, which guarantees updated data. </li>
 * <li>A <i>terminate()</i> function. This lets external code know that the command has completed executing.
 * Local ProtocolZero commands will automatically terminate when the relevant response is read from the server,
 * so the user need only to worry about the overall termination. </li>
 * <li>[Optional] A <i>toString()</i> function, which will format the command output on termination.</ul>
 */
abstract public class ProtocolOne extends Command{
		
	CommandTerminationListener listener;
	List<Command> children;
	
	protected ProtocolOne(CommandType whatami){
		super(false,whatami);
		children = Collections.synchronizedList(new ArrayList<Command>());
	}
	
	protected ProtocolOne(CommandType whatami,long period){
		super(false,whatami,period);
		children = Collections.synchronizedList(new ArrayList<Command>());
	}
	
	protected ProtocolOne(CommandType whatami,long period,long delay){
		super(false,whatami,period,delay);
		children = Collections.synchronizedList(new ArrayList<Command>());
	}
	
	public void terminate() {
		finaltime = System.currentTimeMillis()-inittime;
		running=false;
		comlist.commandTerminated(this);
		executor.notifyCommandTerminationListener(this);
	}
	
	
	public void setExecutor(Entity c){
		waitForTerminate(); //client switch is sync
							//guarantee that only returned commands are ones explicitly called by client
		executor = c;
		id = executor.incrementCommandID();
		addChildren();
		for(int i=0; i<children.size(); i++){
			children.get(i).setExecutor(executor);
		}
	}
	
	public boolean hasTerminated(){return !running;}
	
	/**
	 * P1 command should specify all nested ProtocolZero commands here.
	 */
	abstract protected void addChildren();
	
	public void pause(){
			if(!verifyClient())
				System.err.println("error: must attach client before pausing command " + this.toString());
			paused = true;
			for(int i=0; i<children.size(); i++)
				if(children.get(i).periodic)
					children.get(i).pause();
	}
	
	public void resume(){
		if(periodic){
			if(!verifyClient())
				System.err.println("error: must attach client before pausing command " + this.toString());
			paused = false;
			for(int i=0; i<children.size(); i++)
				if(children.get(i).periodic)
					children.get(i).resume();
		}
		else
			System.err.println("Can't resume atomic command " + this.toString());
	}
}
