package edu.rochester.cs.quagents3.client;
import java.io.Writer;
import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;

/**
 * Class which represents a Radar command.
 */
public class Radar extends ProtocolZero{
	private static final String opcode = "ra";
	
	private final double range;
	
	//data is stored in parallel lists
	private List<Integer> ids;
	private List<String> typenames;
	private List<Double> distance;
	private List<Double[]> facing; 
	private int numentitiesfound;
	
	public Radar(double range, priority_t priority){
		super(true,CommandType.RADAR,priority);
		this.range = range;
		ids = new LinkedList<Integer>();
		typenames = new LinkedList<String>();
		distance = new LinkedList<Double>();
		facing = new LinkedList<Double[]>();
	}
	public Radar(double range){
		super(true,CommandType.RADAR);
		this.range = range;
		ids = new LinkedList<Integer>();
		typenames = new LinkedList<String>();
		distance = new LinkedList<Double>();
		facing = new LinkedList<Double[]>();
	}
	public Radar(double range, priority_t priority,long period){
		super(true,CommandType.RADAR,period,priority);
		this.range = range;
		ids = new LinkedList<Integer>();
		typenames = new LinkedList<String>();
		distance = new LinkedList<Double>();
		facing = new LinkedList<Double[]>();
	}
	public Radar(double range,long period){
		super(true,CommandType.RADAR,period);
		this.range = range;
		ids = new LinkedList<Integer>();
		typenames = new LinkedList<String>();
		distance = new LinkedList<Double>();
		facing = new LinkedList<Double[]>();
	}
	public Radar(double range, priority_t priority,long period,long delay){
		super(true,CommandType.RADAR,period,delay,priority);
		this.range = range;
		ids = new LinkedList<Integer>();
		typenames = new LinkedList<String>();
		distance = new LinkedList<Double>();
		facing = new LinkedList<Double[]>();
	}
	public Radar(double range,long period,long delay){
		super(true,CommandType.RADAR,period,delay);
		this.range = range;
		ids = new LinkedList<Integer>();
		typenames = new LinkedList<String>();
		distance = new LinkedList<Double>();
		facing = new LinkedList<Double[]>();
	}
	
	
	protected void setOp(){
		op = Client.convertPriority(priority) + " " + opcode + " " + id+
		" " + range;
	}
	
	/**
	 * Allows access to parameter data.
	 * @return Range specified to server.
	 */
	public double getRange(){return range;}
	
	/**
	 * Keeps track of the total number of entities found by the radar.
	 * @return The number of entities found by the radar.
	 */
	public int getNumberEntitiesFound(){
		return numentitiesfound;}
	
	/**
	 * Given index, return entity ID.
	 * @param index Index of entity. Entities are indexed so that the first 
	 * entity found is at the front of the list, and subsequent entities are added to the back of the list.
	 * @return Entity ID corresponding to index.
	 */
	public Integer getID(int index){;
		return ids.get(index);
	}
	
	/**
	 * Given entity index, returns the classtype of entity.
	 * @param index Index of entity. Entities are indexed so that the first 
	 * entity found is at the front of the list, and subsequent entities are added to the back of the list.
	 * @return Entity classtype corresponding to index.
	 */
	public String getClasstype(int index){
		return typenames.get(index);
	}
	
	/**
	 * Given entity index, returns the distance from Quagent to entity.
	 * @param index Index of entity. Entities are indexed so that the first 
	 * entity found is at the front of the list, and subsequent entities are added to the back of the list.
	 * @return Distance to entity corresponding to index.
	 */
	public Double getDistance(int index){
		return distance.get(index);
	}
	
	/**
	 * Finds the index corresponding to a specified entity ID. This should be unique,
	 * so we only need to return an int.
	 * @usage May be used as "CanSee" by calling this. If nonzero can see, else cannot see.
	 * @param entityid Entity ID to search for.
	 * @return If found, index which corresponds to the entity ID. Else -1.
	 */
	public int getIndex(int entityid){
		for(int i=0; i<ids.size(); i++)
			if(ids.get(i)==entityid)
				return i;
		return -1;
	}
	
	/**
	 * Finds the indexes corresponding to a specified classtype.
	 * @param classtype Classtype to search for.
	 * @return List of all entity indexes corresponding to the classtype.
	 */
	public List<Integer> getIndex(String classtype){
		List<Integer> list = new LinkedList<Integer>();
		for(int i=0; i<typenames.size(); i++)
			if(typenames.get(i).equals(classtype))
				list.add(i);
		return list;
	}
	
	/**
	 * Gets a list of all entities (identified by entity index) that exist on a circle of radius <i>range</i>.
	 * @param range Radius of ring.
	 * @return List of all entity indexes who exist distance <i>range</i> from the Quagent.
	 */
	public List<Integer> getAllAtRange(double range){
		List<Integer> list = new LinkedList<Integer>();
		for(int i=0; i<distance.size(); i++)
			if(distance.get(i).equals(range))
				list.add(i);
		return list;
	}
	
	/**
	 * Gets the direction vector change between the Quagent's current facing and the entity.
	 * E.g., the quagent would need to rotate(yaw,pitch) to face entity.
	 * @param index Index of entity. Entities are indexed so that the first 
	 * entity found is at the front of the list, and subsequent entities are added to the back of the list.
	 * @return Double[] of yaw and pitch to target, in form [YAW,PITCH]. 
	 * <br/> Yaw is in degrees left of Quagent, pitch is in degrees down from Quagent's facing.
	 */
	public Double[] getFacing(int index){
		return facing.get(index);
	}
	
	/**
	 * Gets the data string unique to the specified entity index. This takes the form of a formatted string
	 * containing classname, entityid, distance to, yaw, and pitch.
	 * @param index Index of entity. Entities are indexed so that the first 
	 * entity found is at the front of the list, and subsequent entities are added to the back of the list.
	 * @return Formatted data string for entity at index.
	 */
	public String getEntityDataString(int index){
		return "entity: " + typenames.get(index) + " of entityid: " + ids.get(index) +
		" is at distance: " + distance.get(index) + " yaw: " + facing.get(index)[0] + " pitch: " + facing.get(index)[1];
	}
	
	/**
	 * Gets all data strings for all entity indexes. 
	 * @return Formatted string with all data strings separated by newlines.
	 */
	//this may be very slow with a _large_ number of entities
	public String getAllEntityDataStrings(){
		String rtn = "";
		for(int i=0;i<ids.size();i++){
			rtn += getEntityDataString(i) + '\n';
		}
		return rtn;
	}
	
	public void setData(String params) {
		datastring += params + '\n';
		numentitiesfound++;
		paramscan = new Scanner(params);
		ids.add(paramscan.nextInt());
		typenames.add(paramscan.next());
		distance.add(paramscan.nextDouble());
		Double[] temp = {paramscan.nextDouble(), paramscan.nextDouble()};
		facing.add(temp);
	}
}
