package edu.rochester.cs.quagents3.client;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Timer;

/**
 * Class which handles the periodic issuing of commands and queries.
 */
public class Updater extends Thread{
	private List<Command> ops;
	private ArrayList<Timer> scheduler;
	private boolean running;
	
	public Updater(){
		ops = Collections.synchronizedList(new ArrayList<Command>());
		scheduler = new ArrayList<Timer>();
		running = false;
	}
	
	/**
	 * Inserts the command into a new timer, which will run the command
	 * according to that object's <i>period</i> and <i>delay</i> fields.
	 * @param com Command to be issued.
	 */
	//TODO: Allow for non-repeating commands?
	public void insert(Command com){
		if(contains(com)) return; //don't want duplicate functions
		Timer t = new Timer();
		scheduler.add(t);
		ops.add(com);
		if(running)
			t.scheduleAtFixedRate(com, com.getDelay(),com.getPeriod());
	}
	
	/**
	 * Returns the command identified by command ID.
	 * @usage Used to keep track of command objects, in order to access
	 * response data.
	 * @param commandid Command ID of target command.
	 * @return Command identitfied by <i>commandid</i>.
	 */
	public Command getCommand(long commandid){
		for(int i=0; i<ops.size(); i++)
			if(ops.get(i).getID()==commandid)
				return ops.get(i);
		return null; //if not found
	}
	
	/**
	 * Checks to see if a certain command already exists within the updater.
	 * @usage Used by <i>insert()</i> to prevent duplicate commands.
	 * @param com Command to be tested.
	 * @return True if this contains the command, false if not.
	 */
	private boolean contains(Command com){
		for(int i=0; i<ops.size(); i++)
			if(com.compareTo(ops.get(i))==0)
				return true;
		return false;
	}

	/**
	 * Allows the user to remove a command from auto-updating.
	 * @param com Command to remove.
	 * @return Removed command.
	 */
	public Command remove(Command com){
		for(int i=0; i<ops.size(); i++){
			if(com.compareTo(ops.get(i))==0){
				//we have an equal command, remove
				scheduler.get(i).cancel();
				scheduler.remove(i);
				return ops.remove(i);
			}
		}
		System.err.print("error: command not found");
		return null; //not found, return nothing
	}
	
	
	public Command remove(long commandid){
		for(int i=0; i<ops.size(); i++){
			if(ops.get(i).getID()==commandid){
				//we have an equal command, remove
				scheduler.get(i).cancel();
				scheduler.remove(i);
				return ops.remove(i);
			}
		}
		System.err.print("error: command not found");
		return null; //not found, return nothing
	}
	
	public void run(){
		running = true;
		for(int i=0; i<ops.size(); i++)
			scheduler.get(i).scheduleAtFixedRate(ops.get(i), ops.get(i).getDelay(), ops.get(i).getPeriod());
	}
}
