/*
 * File: QuagentController.java
 * Creator: George Ferguson
 * Created: Thu May 31 09:37:01 2012
 * Time-stamp: <Fri May 17 09:25:07 EDT 2013 ferguson>
 */

import java.io.*;
import java.net.*;

/**
 * Base class for implementing agents that control quagents (bots) in the
 * Quagents3 environment.
 * This class provides the core core connection and messaging capabilities,
 * but no UI and no behavior of its own.
 * <p>
 * This class of agent is single-threaded. Once initialized, it sits in a
 * Perceive-Behave loop driven by the receipt of messages from
 * the Quagent server. This class doesn't actually do anything, but
 * subclasses can by overriding the appropriate methods to implement
 * behaviors.
 */
public class QuagentController implements Runnable {

    /**
     * Construct and return a new QuagentController connected to the
     * Quagents3 server.
     */
    public QuagentController() throws IOException {
	super();
	initSocket();
    }

    protected Socket socket;
    protected BufferedReader in;
    protected PrintWriter out;

    /**
     * Initialize this QuagentController's connection to the server.
     */
    protected void initSocket() throws IOException {
	socket = new Socket("localhost", 6000);
	in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
	out = new PrintWriter(socket.getOutputStream(), true);
    }

    //
    // Properties
    //

    /**
     * This QuagentController's qbot's name.
     */
    protected String name;

    /**
     * Return the name of this QuagentController's qbot.
     */
    public String getName() {
	return name;
    }

    /**
     * Set the name of this QuagentController's qbot.
     * This method will only work during the agent's initialization phase
     * (i.e., before calling ready()).
     */
    public void setName(String name) {
	this.name = name;
    }

    //
    // Client-server IO
    //

    /**
     * Send the given string to the server (as well as logging to stdout).
     */
    protected void sendString(String str) {
	System.out.println("sending: " + str);
	out.println(str);
    }

    /**
     * Read and return a message from the Quagent server, blocking
     * until one is available.
     */
    protected String readMessage() throws IOException {
	return in.readLine();
    }

    //
    // Quagent Protocol commands
    //

    /**
     * Enumeration of possible values for the <q>scheduling</q> parameter
     * of Quagent protocol commands.
     */
    public enum When {
	NOW	('n'),
	THEN	('t'),
	REPLACE	('r');
	private When(char code) {
	    this.code = code;
	}
	protected char code;
	public char getCode() {
	    return code;
	}
    }

    /**
     * Send the given command to the Quagents3 server following the
     * protocol specification.
     */
    protected void sendCommand(When when, String op, int id, String args) {
	String cmd = String.format("%c %s %d %s", when.getCode(), op, id, args);
	sendString(cmd);
    }

    /**
     * Send the given command (opcode and args) to the Quagents3 server
     * following the protocol specification. The command is sent for
     * execution <q>now</q> and we maintain a counter for the command id
     * parameter.
     */
    protected void sendCommand(String op, String args) {
	sendCommand(When.NOW, op, cmdId++, args);
    }

    protected int cmdId = 1;

    /**
     * Send the given command (which takes no arguments) to the Quagents3
     * server.
     */
    protected void sendCommand(String op) {
	sendCommand(op, "");
    }

    //
    // Agent behavior
    //

    /**
     * Flag used in agent main loop to indicate that this QuagentController
     * should stop execution.
     */
    protected boolean stopped = false;

    /**
     * Tells this Quagent agent to stop as soon as possible.
     */
    public void stop() {
	stopped = true;
    }

    /**
     * Main Runnable method for a QuagentController: init, ready,
     * then perceive-behave until stopped or exception.
     */
    public void run() {
	try {
	    init();
	    ready();
	    while (!stopped) {
		perceive();
		if (!stopped) {
		    behave();
		}
	    }
	} catch (EOFException ex) {
	} catch (IOException ex) {
	    System.err.println(ex);
	}
	try {
	    shutdown();
	} catch (IOException ex) {
	}
    }

    /**
     * Called when an QuagentController starts execution.
     */
    protected void init() throws IOException {
	sendString("name " + name);
    }

    /**
     * Called when an QuagentController is ready to go live.
     */
    protected void ready() throws IOException {
	sendString("ready");
    }

    /**
     * Called to update an QuagentController's perceptual state.
     */
    protected void perceive() throws IOException {
	String msg = readMessage();
	System.out.println(msg);
    }

    /**
     * Called to allow an QuagentController to think and act based on
     * its perceptions (or not!).
     */
    protected void behave() {
    }

    /**
     * Shutdown this QuagentController, which involves closing
     * its IO streams.
     */
    public void shutdown() throws IOException {
	in.close();
	out.close();
    }

    //
    // Testing
    //

    /**
     * Main method for QuagentController: connects to the server,
     * makes the bot appear, then sits reading messages until killed.
     */
    public static void main(String[] argv) throws IOException, InterruptedException {
	QuagentController quagent = new QuagentController();
	String name = argv.length > 0 ? argv[0] : "QuagentController";
	quagent.setName(name);
	quagent.run();
    }

}
